/*********************************************************************
* STUFF - This program stuffs text into the BIOS keyboard buffer.    *
*   You can use this program to stuff a DOS command into the buffer  *
*   so DOS will execute the command when this program terminates.    *
*                                                                    *
* To compile: cl KEYBRD.C                                            *
*                                                                    *
* Author: Mike Finley, Last Revised: 10/10/91                        *
*********************************************************************/
#include "stdio.h"  /* printf(), puts(), getc(), EOF,               */

/*----Define BIOS data areas for keyboard---------------------------*/
#define KEYB_HEAD   (int  far *)0x0040001A /* Head pointer          */
#define KEYB_TAIL   (int  far *)0x0040001C /* Tail pointer          */
#define KEYB_BUFF   (char far *)0x00400000 /* Start of BIOS segment */
#define KEYB_START  (int  far *)0x00400080 /* Start-of-buffer ptr   */
#define KEYB_END    (int  far *)0x00400082 /* End-of-buffer ptr     */
#define KEYB_LEN    (*KEYB_END-*KEYB_START)/* Length = end-start    */

/*----Define a few local keyboard value constants-------------------*/
#define ASCIIZ          '\0'           /* ASCII 0 terminator byte   */
#define NO_KEYS         -1             /* Buffer empty indicator    */

/*----Booleans------------------------------------------------------*/
#define FALSE           0
#define TRUE            1

/*----Prototypes----------------------------------------------------*/
int ReadAhead( int );
void DeleteKey( void );
int KeysTyped( void );
int GetAKey( void );
int KeybFull( void );
int KeybEmpty( void );
void ClearKeyb( void );
void WriteKeyb( char * );

/*********************************************************************
* ReadAhead() - Reads the iForwardth character in the buffer.        *
*   For example, ReadAhead(0) reads the next available character.    *
*********************************************************************/
int ReadAhead( int iForward )
    {
    int  far *fpiKeyHead = KEYB_HEAD;  /* Offset of last key typed  */
    int  far *fpiKeyTail = KEYB_TAIL;  /* Next available slot       */
    char far *fpcKeyBuff = KEYB_BUFF;  /* Address of keyboard buffer*/
    int  iForwardOffset;               /* Offset of key sought      */
    char iRetChar;                     /* Return value              */

    /*----Check to see if iForwardth character exists---------------*/
    if ( KeysTyped() < (iForward+1) )
        iRetChar = NO_KEYS;
    else
        {
        iForwardOffset = *fpiKeyHead + (iForward << 1) ;
        if ( iForwardOffset >= *KEYB_END )  /* Calc. the offset of  */
            iForwardOffset -= KEYB_LEN;     /* the sought character */
        iRetChar = fpcKeyBuff[iForwardOffset]; /* Get char from buff*/
        }

    return( iRetChar );
    }

/*********************************************************************
* DeleteKey() - Deletes next available key from the keyboard buffer  *
*********************************************************************/
void DeleteKey( void )
    {
    int  far *fpiKeyHead = KEYB_HEAD; /* Offset of oldest character */

    if ( KeysTyped() > 0 )            /* Look for keys in the buffer*/
        if ( ((*fpiKeyHead)+=2) >= *KEYB_END )   /* wrap, if needed */
            *fpiKeyHead = *KEYB_START;
    }

/*********************************************************************
* KeysTyped() - Returns number of keystrokes in the keyboard buffer  *
*********************************************************************/
int KeysTyped( void )
    {
    int  far *fpiKeyHead = KEYB_HEAD;  /* Offset of oldest character*/
    int  far *fpiKeyTail = KEYB_TAIL;  /* Next available slot       */
    int  iRetNum;                      /* Return value              */

    iRetNum = (*fpiKeyTail - *fpiKeyHead) >> 1;    /* (head-tail)/2 */
    if ( iRetNum < 0 )                             /* wraparound    */
        iRetNum += KEYB_LEN >> 1;

    return( iRetNum );
    }

/*********************************************************************
* GetAKey() - Returns the next available character from the buffer.  *
*********************************************************************/
int GetAKey( void )
    {
    int iChar;

    /*----Wait for a character to appear in buffer------------------*/
    while ( (iChar = ReadAhead(0)) == NO_KEYS )
        ;
    DeleteKey();                   /* Remove character from buffer  */

    return( iChar );
    }

/*********************************************************************
* ClearKeyb() - Clears all characters from keyboard buffer by        *
*   the head pointer equal to the tail pointer                       *
*********************************************************************/
void ClearKeyb( void )
    {
    int  far *fpiKeyTail = KEYB_TAIL; /* Next available slot        */
    int  far *fpiKeyHead = KEYB_HEAD; /* Offset of oldest character */
    *fpiKeyHead = *fpiKeyTail;
    }

/*********************************************************************
* WriteKeyb() - Stuffs the null-terminated string TypeMe into the    *
*   the keyboard buffer just as if the user had typed it             *
*********************************************************************/
void WriteKeyb( char *pszTypeMe )
    {
    int  far *fpiKeyTail = KEYB_TAIL; /* Next available slot        */
    char far *fpcKeyBuff = KEYB_BUFF; /* Address of keyboard buffer */

    _disable();              /* No interrupts while we're stuffing  */
    while ( *pszTypeMe != ASCIIZ && !KeybFull() )  /* Stuff 'em!    */
        {
        fpcKeyBuff[(*fpiKeyTail)++] = *pszTypeMe++; /* Stuff char   */
        fpcKeyBuff[(*fpiKeyTail)++] = 0;          /* Stuff scancode */
        if ( *fpiKeyTail == *KEYB_END )           /* Wraparound if  */
            *fpiKeyTail = *KEYB_START;            /* necessary      */
        }
    _enable();               /* Let interrupts through again        */
    }

/*********************************************************************
* KeybFull() - Returns TRUE if the keyboard buffer is full           *
*********************************************************************/
int KeybFull( void )
    {
    int  far *fpiKeyTail = KEYB_TAIL; /* Next available slot        */
    int  far *fpiKeyHead = KEYB_HEAD; /* Offset of oldest character */

    /*----Check for Tail about to meet tail, including wraparound---*/
    if ( (*fpiKeyTail+2) == (*fpiKeyHead) ||
         (*fpiKeyTail == *KEYB_END && *fpiKeyHead == *KEYB_START) )
        return( TRUE );
    else
        return( FALSE );
    }

/*********************************************************************
* KeybEmpty() - Returns TRUE if the keyboard buffer is empty         *
*********************************************************************/
int KeybEmpty( void )
    {
    int  far *fpiKeyTail = KEYB_TAIL; /* Next available slot        */
    int  far *fpiKeyHead = KEYB_HEAD; /* Offset of oldest character */

    /*----Check for Tail equal to head; if they're equal, it's empty*/
    if ( *fpiKeyTail == *fpiKeyHead )
        return( TRUE );
    else
        return( FALSE );
    }

/*********************************************************************
* Main body of program                                               *
*********************************************************************/
int main( int argc, char **argv )
    {
    int i;
    char DosCmd[128] = "";

    /*----First, look for command on DOS command-line---------------*/
    if ( argc > 1 )
        {
        for( i=1; i < argc; i++ )
            {
            if ( i > 1 )
                strcat( DosCmd, " " );
            strcat( DosCmd, argv[i] );
            }
        }
    else
    /*----If command-line empty, ask user for command---------------*/
    {
        printf( "\nType a DOS command (14 char max)"
                " to execute on completion: " );
        gets( DosCmd );
    }

    printf( "Press any key to stuff the command '%s'", DosCmd );
    printf( " and exit..." );
    ClearKeyb();
    GetAKey();

    /*----Stuff the command before exiting--------------------------*/
    WriteKeyb( DosCmd );
    WriteKeyb( "\r" );
    }

